@@ -6,6 +6,8 @@
 # Description : An FTP client for POE
 
 package POE::Component::Client::FTP;
+$POE::Component::Client::FTP::VERSION = '0.24';
+#ABSTRACT: Implements an FTP client POE Component
 
 use strict;
 use warnings;
@@ -18,9 +20,7 @@ use Socket;
 use POE qw(Wheel::SocketFactory Wheel::ReadWrite
 	   Filter::Stream Filter::Line Driver::SysRW);
 
-use vars qw(@ISA @EXPORT $VERSION $poe_kernel);
-
-$VERSION = '0.22';
+use vars qw(@ISA @EXPORT $poe_kernel);
 
 @ISA = qw(Exporter);
 @EXPORT = qw(FTP_PASSIVE FTP_ACTIVE FTP_MANUAL FTP_ASCII FTP_BINARY);
@@ -82,7 +82,7 @@ my $state_map =
 
 		map ( { $_ => \&do_simple_command }
 		      qw{ cd cdup delete mdtm mkdir noop
-			  pwd rmdir site size stat type quit quot }
+			  pwd rmdir site size stat syst type quit quot }
 		    ),
 
 		map ( { $_ => \&do_complex_command }
@@ -96,8 +96,8 @@ my $state_map =
 
     put => { data_flush         => \&handler_put_flushed,
 	     data_error         => \&handler_put_data_error,
-	     
-             put_data           => \&do_put_data,
+
+       put_data           => \&do_put_data,
 	     put_close          => \&do_put_close,
 
 	     data_connected     => \&handler_complex_connected,
@@ -207,7 +207,7 @@ sub spawn {
   # Make sure the user didn't make a typo on parameters
   carp "Unknown parameters: ", join( ', ', sort keys %params )
     if keys %params;
-  
+
   if ( $tlscmd || $tlsdata ) {
     eval {
        require POE::Component::SSLify;
@@ -256,7 +256,7 @@ sub spawn {
 
       events          => { $sender => \%register }
   }, $class;
-  
+
   $self->{session_id} = POE::Session->create (
     inline_states => map_all($state_map, \&dispatch),
 
@@ -296,8 +296,7 @@ sub do_stop {
 
   delete $heap->{cmd_rw_wheel};
   delete $heap->{cmd_sock_wheel};
-  delete $heap->{data_rw_wheel};
-  delete $heap->{data_sock_wheel};
+  clean_up_complex_cmd();
 
   $poe_kernel->alias_remove( $heap->{alias} );
   return;
@@ -311,13 +310,12 @@ sub handler_cmd_input {
 
   my $coderef;
 
-  $input =~ s/^(\d\d\d)(.?)//o;
-  my ($code, $more) = ($1, $2);
+  my ($code, $more) = parse_cmd_string(\$input);
 
   $input =~ s/^ // if defined $more && $more eq "-";
 
   $heap->{ftp_message} .= "$input\n";
-    
+
   return unless defined $code && defined $more && $more eq " ";
 
   $heap->{ftp_message} =~ s/\s+$//;
@@ -472,8 +470,7 @@ sub handler_put_flushed {
   elsif ($heap->{data_suicidal}) {
     warn "killing suicidal socket" . $heap->{data_rw_wheel}->get_driver_out_octets() if DEBUG;
 
-    delete $heap->{data_sock_wheel};
-    delete $heap->{data_rw_wheel};
+    clean_up_complex_cmd();
     $heap->{data_suicidal} = 0;
 
     send_event("put_closed",
@@ -491,8 +488,7 @@ sub handler_put_data_error {
   send_event( "put_error", $error,
 	      $heap->{complex_stack}->{command}->[1] );
 
-  delete $heap->{data_sock_wheel};
-  delete $heap->{data_rw_wheel};
+  clean_up_complex_cmd();
   goto_state("ready");
   return;
 }
@@ -569,7 +565,7 @@ sub handler_init_success {
   my $status = substr($input, 0, 3);
   my $line = substr($input, 4);
 
-  send_event( "connected", 
+  send_event( "connected",
 	      $status, $line );
 
 
@@ -810,8 +806,6 @@ sub handler_complex_success {
     command($heap->{complex_stack}->{command});
   }
   else {
-    send_event( $heap->{complex_stack}->{command}->[0] . "_done",
-                $heap->{complex_stack}->{command}->[1] );
     goto_state("ready");
   }
   return;
@@ -828,9 +822,7 @@ sub handler_complex_failure {
 	      $status, $line,
 	      $heap->{complex_stack}->{command}->[1] );
 
-  delete $heap->{data_sock_wheel};
-  delete $heap->{data_rw_wheel};
-
+  clean_up_complex_cmd();
   goto_state("ready");
   return;
 }
@@ -868,7 +860,7 @@ sub handler_complex_preliminary {
 # data connection established
 sub handler_complex_connected {
   my ($kernel, $heap, $socket) = @_[KERNEL, HEAP, ARG0];
-  
+
   $heap->{data_rw_wheel} = POE::Wheel::ReadWrite->new(
     Handle       => $socket,
     Filter       => $heap->{filters}->{ $heap->{complex_stack}->{command}->[0] },
@@ -893,8 +885,7 @@ sub handler_complex_connect_error {
   send_event( $heap->{complex_stack}->{command}->[0] . "_error", $error,
 	      $heap->{complex_stack}->{command}->[1] );
 
-  delete $heap->{data_sock_wheel};
-  delete $heap->{data_rw_wheel};
+  clean_up_complex_cmd();
   goto_state("ready");
   return;
 }
@@ -913,9 +904,11 @@ sub handler_complex_list_data {
 sub handler_complex_list_error {
   my ($kernel, $heap, $input) = @_[KERNEL, HEAP, ARG0];
   warn "error: complex_list: $input" if DEBUG;
+  send_event( $heap->{complex_stack}->{command}->[0] . "_done",
+	      $heap->{complex_stack}->{command}->[1] );
 
-  delete $heap->{data_sock_wheel};
-  delete $heap->{data_rw_wheel};
+  clean_up_complex_cmd();
+  dequeue_complex_cmd();
   return;
 }
 
@@ -960,8 +953,15 @@ sub dequeue_event {
 # if active session knows how to handle this event, dispatch it to them
 # if not, enqueue the event
 sub dispatch {
-  my ($kernel, $heap, $state) = @_[KERNEL, HEAP, STATE];
+  my ($kernel, $heap, $state, $input) = @_[KERNEL, HEAP, STATE, ARG0];
 
+  if ($state eq 'cmd_input' && defined $heap->{data_rw_wheel} ) {
+      my ($code, $msg) = parse_cmd_string(\$input);
+      if ( substr($code, 0, 1) == 2 ) {
+        enqueue_complex_cmd(@_);
+        return;
+      }
+  }
   my $coderef = ( $state_map->{ $heap->{state} }->{$state} ||
 		  $state_map->{global}->{$state} ||
 		  \&enqueue_event );
@@ -1011,10 +1011,10 @@ sub command {
     unshift @{$heap->{stack}}, [ $command, @$cmd_args ];
 
     $command = shift( @$cmd_args ) if $command eq "QUOT";
- 
+
     $command = $command_map{$command} || $command;
     my $cmdstr = join( ' ', $command, @$cmd_args ? @$cmd_args : () );
- 
+
     warn ">>> $cmdstr\n" if DEBUG_COMMAND;
 
     $heap->{cmd_rw_wheel}->put($cmdstr);
@@ -1062,14 +1062,53 @@ sub establish_data_conn {
   return;
 }
 
+sub parse_cmd_string {
+  my $string = shift;
+  $$string =~ s/^(\d\d\d)(.?)//o;
+  my ($code, $more) = ($1, $2);
+  return ($code, $more);
+}
+
+sub clean_up_complex_cmd {
+  my $heap = $poe_kernel->get_active_session()->get_heap();
+  delete $heap->{data_sock_wheel};
+  delete $heap->{data_rw_wheel};
+  $heap->{complex_stack} = {};
+}
+
+sub enqueue_complex_cmd {
+  my $heap = $poe_kernel->get_active_session()->get_heap();
+  warn "enqueue_complex_cmd $_[STATE]" if DEBUG;
+  $heap->{pending_complex_cmd} = \@_;
+}
+
+sub dequeue_complex_cmd {
+  my $heap = $poe_kernel->get_active_session()->get_heap();
+  return unless $heap->{pending_complex_cmd};
+
+  my $state = $heap->{pending_complex_cmd}->[STATE];
+  warn "dequeue_complex_cmd $state" if DEBUG;
+  my @event = @{ $heap->{pending_complex_cmd} };
+  $heap->{pending_complex_cmd} = undef;
+  dispatch( @event );
+}
+
 1;
 
 __END__
 
+=pod
+
+=encoding UTF-8
+
 =head1 NAME
 
 POE::Component::Client::FTP - Implements an FTP client POE Component
 
+=head1 VERSION
+
+version 0.24
+
 =head1 SYNOPSIS
 
   use POE::Component::Client::FTP;
@@ -1143,6 +1182,68 @@ L<http://www.wush.net/poe/ftp>
 POE::Component::Client::FTP is a L<POE> component that implements an non-blocking FTP client.
 One C<spawns> an FTP poco from within one's own POE session, asking to receive particular events.
 
+=for Pod::Coverage        DEBUG
+       DEBUG_COMMAND
+       DEBUG_DATA
+       EOL
+       FTP_ACTIVE
+       FTP_ASCII
+       FTP_BINARY
+       FTP_MANUAL
+       FTP_PASSIVE
+       clean_up_complex_cmd
+       command
+       dequeue_complex_cmd
+       dequeue_event
+       dispatch
+       do_complex_command
+       do_init_start
+       do_login_send_password
+       do_login_send_username
+       do_put
+       do_put_close
+       do_put_data
+       do_rename
+       do_send_authtls
+       do_send_pbsz
+       do_send_prot
+       do_set_attribute
+       do_simple_command
+       do_stop
+       enqueue_complex_cmd
+       enqueue_event
+       establish_data_conn
+       goto_state
+       handler_authtls_failure
+       handler_authtls_success
+       handler_cmd_error
+       handler_cmd_input
+       handler_complex_connect_error
+       handler_complex_connected
+       handler_complex_failure
+       handler_complex_flushed
+       handler_complex_list_data
+       handler_complex_list_error
+       handler_complex_preliminary
+       handler_complex_success
+       handler_init_connected
+       handler_init_error
+       handler_init_success
+       handler_login_failure
+       handler_login_success
+       handler_pbsz_prot_failure
+       handler_pbsz_prot_success
+       handler_put_data_error
+       handler_put_flushed
+       handler_rename_failure
+       handler_rename_intermediate
+       handler_rename_success
+       handler_simple_failure
+       handler_simple_success
+       map_all
+       parse_cmd_string
+       send_event
+
 =head1 CONSTRUCTOR
 
 =over
@@ -1198,48 +1299,50 @@ These are commands which the poco will accept events for:
 
 =over
 
-=item cd [path]
+=item C<cd [path]>
 
-=item cdup
+=item C<cdup>
 
-=item delete [filename]
+=item C<delete [filename]>
 
-=item dir
+=item C<dir>
 
-=item get [filename]
+=item C<get [filename]>
 
-=item ls
+=item C<ls>
 
-=item mdtm [filename]
+=item C<mdtm [filename]>
 
-=item mkdir [dir name]
+=item C<mkdir [dir name]>
 
-=item mode [active passive]
+=item C<mode [active passive]>
 
-=item noop
+=item C<noop>
 
-=item pwd
+=item C<pwd>
 
-=item rmdir [dir name]
+=item C<rmdir [dir name]>
 
-=item site [command]
+=item C<site [command]>
 
-=item size [filename]
+=item C<size [filename]>
 
-=item stat [command]
+=item C<stat [command]>
 
-=item type [A|I]
+=item C<syst>
 
-=item quit
+=item C<type [A|I]>
 
-=item quot [command]
+=item C<quit>
 
-=item put_data
+=item C<quot [command]>
 
-After receiving a put_connected event you can post put_data events to send 
+=item C<put_data>
+
+After receiving a put_connected event you can post put_data events to send
 data to the server.
 
-=item put_close
+=item C<put_close>
 
 Closes the data connection.  put_closed will be emit when connection is flushed
 and closed.
@@ -1326,10 +1429,15 @@ LocalAddr in the constructor and it all works fine.
 
 Please report any other bugs through C<bug-POE-Component-Client-FTP@rt.cpan.org>
 
-=head1 LICENSE
+=head1 AUTHOR
 
-Copyright (c) 2002 Michael Ching. 
-Copyright (c) 2008 Chris Williams.
+Michael Ching <michaelc@wush.net>
 
-This program is free software; you can redistribute it and/or modify it under the same terms as
-Perl itself.
+=head1 COPYRIGHT AND LICENSE
+
+This software is copyright (c) 2017 by Michael Ching and Chris Williams.
+
+This is free software; you can redistribute it and/or modify it under
+the same terms as the Perl 5 programming language system itself.
+
+=cut
